#include "ptheap.h"
#include "qsbase.h"
#include "qsint.h"
#include "stdlib.h"
#include "qsmodext.h"

#define PTH_SIGN         0x49485450          // PTHI
#define PTH_DEBUG        1

typedef struct {
   u32t         sign;
   u32t         nblk,
             nblk64k,  ///< blocks in 64k block
                n64k,
               bsize;  ///< size of the block
   u32t      *bitmap;
   u32t         hint;
   u32t          pid;
   u32t        memfl;
   u8t          **lp;  ///< local process pointers
   u8t          **gp;  ///< system arena pointers
} pthheap;

#define instance_ret(inst,err)       \
   pthheap *inst = (pthheap*)hh;     \
   if (inst->sign!=PTH_SIGN) return err;

#define instance_void(inst)          \
   pthheap *inst = (pthheap*)hh;     \
   if (inst->sign!=PTH_SIGN) return;

qserr pth_init(u32t blocksize, u32t limit, u32t memflags, ptheap *result) {
   u32t nblk64k,
           n64k,
           size;
   u32t   local = memflags & QSMA_LOCAL ? mod_getpid() : 0;
   pthheap *res;

   if (!limit || (memflags & ~(QSMA_LOCAL|QSMA_READONLY))) return E_SYS_INVPARM;
   if (blocksize>_32KB) return E_SYS_TOOLARGE;
   // align it for the bitmap
   limit   = Round32(limit);
   // blocks in 64k block
   nblk64k = _64KB / blocksize;
   // number of 64k blocks for the heap
   n64k    = limit / nblk64k;
   // there is no "limit" align up since its size can be assumed by caller
   if (limit%nblk64k) n64k++;
   // allocate heap structs (in the system memory!)
   size    = sizeof(u8t*)*n64k;
   res     = (pthheap*)hlp_memalloc(sizeof(pthheap) + sizeof(u32t)*(limit>>5) +
                                    size + (local?size:0),
                                    QSMA_RETERR|memflags&QSMA_READONLY);
   if (!res) return E_SYS_NOMEM;

   res->sign    = PTH_SIGN;
   res->bitmap  = (u32t*)(res + 1);
   res->gp      = (u8t**)((u8t*)res->bitmap + size);
   res->lp      = local ? (u8t**)((u8t*)res->gp + size) : 0;
   res->nblk    = limit;
   res->nblk64k = nblk64k;
   res->n64k    = n64k;
   res->bsize   = blocksize;
   res->hint    = 0;
   res->memfl   = memflags;
   res->pid     = local;
   *result      = res;
   return 0;
}

qserr pth_fini(ptheap hh, int freemem) {
   u32t ii;
   instance_ret(ph,E_SYS_INVOBJECT);
   // freemem requires process match!
   if (freemem && ph->pid && mod_getpid()!=ph->pid) return E_SYS_PRIVATEHANDLE;
   ph->sign = 0;

   for (ii=0; ii<ph->nblk; ii++) {
      if (ph->pid && ph->gp[ii]) {
         // unmap should be here
         _throwmsg_("not finished!");
         ph->gp[ii] = 0;
      }
      if (freemem)
         if (ph->lp && ph->lp[ii]) hlp_memfree(ph->lp[ii]); else
            if (ph->gp[ii]) hlp_memfree(ph->gp[ii]);
   }
   hlp_memfree(ph);
   return 0;
}

u32t pth_alloc(ptheap hh) {
   u32t  *rp, res;
   instance_ret(ph,0);

   mt_swlock();
   rp  = ph->bitmap + ph->hint;
   if (*rp==FFFF) rp = memchrnd(rp, FFFF, (ph->nblk>>5) - ph->hint);
   if (rp) {
      u32t bit = bsf32(!*rp);
      _bitset(rp, bit);
      res = rp - ph->bitmap;
      if (res>ph->hint) ph->hint = res;
      // 1 based
      res = (res<<5) + bit + 1;
   } else
      res = 0;
   mt_swunlock();
   return res;
}

/// get/free action. is res==0 - this is free
static qserr getindex(ptheap hh, u32t index, void **res) {
   qserr rc;
   instance_ret(ph,E_SYS_INVOBJECT);
   if (res) *res = 0;
   // this can be often, exit without lock
   if (!index) return E_SYS_INVPARM;

   mt_swlock();
   if (ph->nblk>index) rc = E_SYS_INVPARM; else {
      index--;
      rc = (res ? _bittest(ph->bitmap[index>>5], index&0x1F) :
                 _bitreset(ph->bitmap + (index>>5), index&0x1F))
                ? 0 : E_SYS_NONINITOBJ;
      if (!rc) {
         u32t block = index / ph->nblk64k;
         u8t  **ptr = 0;
         // process owned heap
         if (ph->pid) {
            if (ph->pid == mod_getpid() || !(sys_pagemode() & QPGM_VMA)) 
               ptr = ph->lp + block;
            else {
               // map should be here
               _throwmsg_("not finished!");
            }
         } else
            ptr = ph->gp + block;

         if (ptr) {
            void *dst;
            /* real alloc of a new storage, there is no QSMA_RETERR here!
               only required space is allocated (it can be QSMA_READONLY) */
            if (!*ptr && res) {
               u32t asize = (block < ph->n64k-1 ? ph->nblk64k :
                            (ph->nblk - (ph->nblk64k*ph->n64k-1))) * ph->bsize;
               *ptr = (u8t*)hlp_memalloc(asize, ph->memfl);
#if PTH_DEBUG>0
               log_it(3, "heap %X, block %u/%u, size %u = %X\n", ph, block,
                  ph->n64k, asize, *ptr);
#endif
            }
            
            if (*ptr) {
               index -= block * ph->nblk64k;
               dst    = *ptr + index * ph->bsize;
               /* clear block during free (we can`t clear it at init because
               there is no real alloc in it and can`t in get because then we need
               another one bit about "inited" state) */
               if (res) *res = dst; else
                  memset(dst, 0, ph->bsize);
            }
            // update search hint
            if (!res && ph->hint>index>>5) ph->hint = index>>5;
         }
      }
   }
   mt_swunlock();
   return rc;
}

void *pth_get(ptheap hh, u32t index) {
   void *res;
   if (getindex(hh, index, &res)) return 0;
   return res;
}

qserr pth_free(ptheap hh, u32t index) {
   return getindex(hh, index, 0);
}
