//
// QSINIT EFI loader
// main() function
//
#include "qsloader.h"
#include "efilib.h"
#include "pe.h"
#include "seldesc.h"

#define RESERVE_DESCRIPTORS    256
#define PRINT_MEM_TABLE        0
#define XCPT_STACK_SIZE        (48*1024)

#pragma pack(1)
typedef struct {
   UINT16                limit;
   EFI_PHYSICAL_ADDRESS   base;
} GDT64;
#pragma pack()

void ReadGDTR       (GDT64 *Gdt);
void WriteGDTR      (GDT64 *Gdt);
void ReadIDTR       (GDT64 *Gdt);
void SaveSegRegs    (void);
void RestSegRegs    (void);

EFI_HANDLE              SelfImg = 0;
static GDT64           SavedGDT,
                         CurGDT,
                         CurIDT;
u64t                   SavedCr4;
u32t*                     lapic = 0;       // non-zero only in MT mode
u32t                   SavedSpr = 0;
EFI_PHYSICAL_ADDRESS    GDTCopy = 0;       // own GDT :)
UINT32                  GDTSize = 0,       // size of GDT in bytes
                      GDTLowest = 0,       // lowest usable selector
                   GDTCopyPages = 0;       // # of AllocPages to free
EFI_PHYSICAL_ADDRESS      QSMem = 0;       // memory for QSINIT
UINTN                   QSPages = 0;       // size of it in pages
EFI_LOADED_IMAGE   *LoadedImage = 0;
EFI_DEVICE_PATH     *DevicePath = 0;
EFI_HANDLE         RunModHandle = 0;
CHAR16               *RunModOpt = 0;
UINT32                  UGAOnly = 0;
UINT32                 SafeMode = 0;
UINT32                   Mem640 = 0;       // 1Mb actual memory size
char                    *SetEnv = 0;
UINT32                SetEnvPos = 0;
struct gate64_s      *SavedIDTv = 0;
void                 *XcptStack = 0;
extern u64t        trap_table[];
extern CHAR16          *DevPath;

static UINT8      int_used[256];           // "vector was changed" flag
char               LdiName[LDI_NAME_LEN] = {0};
char               IniName[LDI_NAME_LEN] = {0};

#if PRINT_MEM_TABLE > 0
static CHAR16  *MemTypeStr[EfiMaxMemoryType]  = {
            L"reserved  ",
            L"LoaderCode",
            L"LoaderData",
            L"BS_code   ",
            L"BS_data   ",
            L"RT_code   ",
            L"RT_data   ",
            L"available ",
            L"Unusable  ",
            L"ACPI_recl ",
            L"ACPI_NVS  ",
            L"MemMapIO  ",
			L"MemPortIO ",
			L"PAL_code  "
    };
#endif

extern void timer_irq(void);
extern void iretq_addr(void);

EFI_STATUS OpenProtocolByHandle(EFI_HANDLE handle, EFI_GUID *protocol,
                                void **interface)
{
   return (BS->OpenProtocol(handle, protocol, interface, SelfImg, 0,
           EFI_OPEN_PROTOCOL_BY_HANDLE_PROTOCOL));
}

void Shutdown(CHAR16 *msg) {
   mfs_term();
   vio_term();
   // here we get back EFI console
   if (msg) Print(msg);
   dsk_term();
   // restore GDT
   if (GDTCopy) {
      GDT64  cGDT;
      ReadGDTR(&cGDT);
      Print (L"GDT base: %X, size %X\n", cGDT.base, cGDT.limit+1);
      // if GDT is ours - restore original
      if (cGDT.base==GDTCopy) {
         RestSegRegs();
         WriteGDTR(&SavedGDT);
         Print (L"GDT restored\n");
      }
      // GDT buffer in EFI memory? Free it
      if (GDTCopyPages) BS->FreePages(GDTCopy, GDTCopyPages);
      GDTCopy = 0;
   }
   if (SetEnv) {
      BS->FreePages((EFI_PHYSICAL_ADDRESS)SetEnv, 1);
      SetEnv = 0;
   }
   // free QSINIT memory space if allocated
   if (QSMem) {
      BS->FreePages(QSMem, QSPages);
      QSMem = 0;
   }
}

void ErrorExit(CHAR16 *msg) {
   Shutdown(msg);
   // return to EFI shell/boot sequence
   BS->Exit(SelfImg, EFI_SUCCESS, 0, NULL);
}

int StrWToA(const CHAR16 *src, CHAR8 *dst, UINTN dstlen) {
   if (!dstlen || !src || !dst) return 0;
   while (--dstlen) {
      CHAR16 ch = *src++;
      if (!ch) break;
      if (ch>=256) return 0;
      *dst++ = (CHAR8)ch;
   }
   *dst = 0;
   return 1;
}

/// push string to the 32-bit part environment
int push_env(char *s1) {
   if (!SetEnv) {
      SetEnv    = (char*)PageAlloc(1);
      SetEnvPos = 0;
   }
   if (SetEnv) {
      u32t len = (u32t)strlen(s1);
      if (SetEnvPos + len + 2 <= PAGESIZE) {
         RtCopyMem(SetEnv+SetEnvPos, s1, len+1);
         SetEnvPos += len + 1;
         return 1;
      }
   }
   return 0;
}

void parse_args(void) {
   // primitive command line parser
   if (LoadedImage->LoadOptionsSize) {
      CHAR16  *cmdline = LoadedImage->LoadOptions;
      char    opt[144];
      UINT32   ii, optlen = 0, setvar = 0;

      for (ii=0; ii<=LoadedImage->LoadOptionsSize; ii++) {
         CHAR16 ch = ii<LoadedImage->LoadOptionsSize ? cmdline[ii] : 0;

         if (ch>' ') {
            if (optlen<sizeof(opt)-1) {
               // do not uppercase anything after =
               if (!setvar && ch>='a' && ch<='z') ch-=0x20;
               opt[optlen++] = (char)ch;
               if (ch=='=') setvar = 1;
            }
         } else {
            if (optlen) {
               opt[optlen] = 0;

               if (!strcmpa(opt,"/UGA")) UGAOnly = 1; else
               if (!strcmpa(opt,"/SM")) SafeMode = 1; else
               if (strnicmp(opt,"/BIN:",5)==0 || strnicmp(opt,"/INI:",5)==0) {
                  u32t   len = (u32t)strlen(opt+5);
                  char *dptr = opt[1]=='I' ? IniName : LdiName;

                  if (len>=LDI_NAME_LEN-1) *dptr = 0; else
                     RtCopyMem(dptr, opt+5, len+1);
               } else
               if (setvar && optlen>=4 && opt[0]=='/' && opt[1]=='@' &&
                  opt[2]!='=') push_env(opt+2);

               optlen = 0;
               setvar = 0;
            }
            if (!ch) break;
         }
      }
   }
   if (SetEnv)
      if (!SetEnvPos) { // ???
         BS->FreePages((EFI_PHYSICAL_ADDRESS)SetEnv, 1);
         SetEnv = 0;
      } else {
         // finalize env.block
         SetEnv[SetEnvPos] = 0;
      }
}

void init_memory(void) {
   EFI_STATUS             Status;
   EFI_MEMORY_DESCRIPTOR    *MME,
                      *MemoryMap;
   UINTN               NoEntries;
   UINTN                  MapKey;
   UINTN                DescSize;
   UINT32                DescVer;
   EFI_PHYSICAL_ADDRESS    RcMem = 0;
   UINTN                 RcPages = 0, 
                       MaxContig = 0;
   UINTN               TotalFree = 0;
   UINTN                      ii;

   MemoryMap = LibMemoryMap(&NoEntries, &MapKey, &DescSize, &DescVer);
   if (!MemoryMap) ErrorExit(L"Can not retrieve the current memory map\n");

#if PRINT_MEM_TABLE > 0
   Print(L"Memory Descriptor List:\n\n");
   Print(L"  Type        Start Address     End Address       Attributes      \n");
   Print(L"  ==========  ================  ================  ================\n");
#endif
   MME = MemoryMap;
   /* calculate:
      - number of free pages below 4Gb 
      - max contigous block size
      - actual size of 640k in 1st Mb */
   for (ii=0; ii<NoEntries; ii++) {
      if (MME->PhysicalStart<_1MB && (MME->Type==EfiConventionalMemory ||
         MME->Type==EfiLoaderCode || MME->Type==EfiLoaderData ||
           MME->Type==EfiBootServicesCode || MME->Type==EfiBootServicesData ||
              MME->Type==EfiRuntimeServicesCode || MME->Type==EfiRuntimeServicesData))
      {
         EFI_PHYSICAL_ADDRESS bend = MME->PhysicalStart+(MME->NumberOfPages<<EFI_PAGE_SHIFT);
         if (bend<=0xA0000 && bend>Mem640) Mem640 = (UINT32)bend;
      }
      if (MME->Type==EfiConventionalMemory && MME->PhysicalStart>=_1MB &&
         MME->PhysicalStart<_4GBLL) 
      {
         TotalFree+=MME->NumberOfPages;
         if (MME->NumberOfPages>MaxContig) MaxContig = MME->NumberOfPages;
      }
#if PRINT_MEM_TABLE > 0
      Print(L"  %s  %X  %X  %X\n", MemTypeStr[MME->Type], MME->PhysicalStart,
            MME->PhysicalStart+(MME->NumberOfPages<<EFI_PAGE_SHIFT)-1, MME->Attribute);
#endif
      MME = NextMemoryDescriptor(MME, DescSize);
   }
   // try to allocate block at the top of 4Gb!
   MME = MemoryMap;
   for (ii=0; ii<NoEntries; ii++) {
      if (MME->Type==EfiConventionalMemory && MME->NumberOfPages<<EFI_PAGE_SHIFT>=_32MB &&
         MME->PhysicalStart>=_64MB && MME->PhysicalStart<_4GBLL)
      {
         EFI_ALLOCATE_TYPE target = AllocateAddress;
         RcPages = MME->NumberOfPages;
         RcMem   = MME->PhysicalStart;

         /* if block >1.5Gb or >1/2 of all available memory - use 1/2 of it */
         if (RcPages>_1GB+_512MB>>PAGESHIFT || RcPages>TotalFree/2 ) {
            RcPages = RcPages/2;
            target  = AllocateMaxAddress;
            RcMem   = FFFF & ~PAGEMASK;
         }
         Status = BS->AllocatePages(target, EfiLoaderData, RcPages, &RcMem);
         if (EFI_ERROR(Status)) RcMem = 0; else break;
      }
      MME = NextMemoryDescriptor(MME, DescSize);
   }
   /* failed to allocate at lowest address? then try to use:
      1. 2/3 of max contigous below 4Gb
      2. 32Mb below 4Gb
      3. 32Mb anywhere */
   if (!RcMem) {
      for (ii=0; ii<3; ii++) {
         RcMem   = FFFF & ~PAGEMASK;
         RcPages = ii ? _32MB>>EFI_PAGE_SHIFT : MaxContig*2/3;

         Status  = BS->AllocatePages(ii<2?AllocateMaxAddress:AllocateAnyPages,
            EfiLoaderData, RcPages, &RcMem);
         if (!EFI_ERROR(Status)) break; else RcMem = 0;
      }
   }
   FreePool(MemoryMap);
   // check result
   if (!RcMem || RcPages < _32MB>>EFI_PAGE_SHIFT) ErrorExit(L"No 32Mb of free memory!\n");
   if (RcMem>=_4GBLL) ErrorExit(L"Unable to allocate memory below 4Gb!\n");
   // finally, we got it!
   if (RcMem && RcPages) {
      QSMem   = RcMem;
      QSPages = RcPages;
   }
}


EFI_PHYSICAL_ADDRESS PageAlloc(u32t pages) {
   EFI_PHYSICAL_ADDRESS Addr = FFFF & ~PAGEMASK;
   EFI_STATUS         Status = BS->AllocatePages(AllocateMaxAddress, EfiLoaderData, 
                                                 pages, &Addr);
   if (EFI_ERROR(Status)) ErrorExit(L"Memory allocation error!\n");
   return Addr;
}

u32t XcptOn(void) {
   if (!SafeMode) {
      struct gate64_s *idesc = (struct gate64_s *)CurIDT.base;
      u16t cs64 = GetCodeSelector();
      int    ii = 0;

      IntsOff();
      while (trap_table[ii]!=FFFF64) {
         u64t inta = trap_table[ii];
         if (inta) {
            idesc->g64_sel    = cs64;
            idesc->g64_ofs0   = (u16t)inta;
            idesc->g64_ofs16  = (u16t)(inta>>16);
            idesc->g64_ofs32  = (u32t)(inta>>32);
            idesc->g64_parms  = 0;
            idesc->g64_access = D_PRES|D_TRAPGATE32;
            // mark it to restore
            int_used[ii] = 1;
         }
         ii++;
         idesc++;
      }
      IntsOn();
      return 1;
   }
   return 0;
}

u32t TimerIrqsOn(u32t* apic, u8t tmr_vector, u8t spr_vector) {
   if (tmr_vector<16 || spr_vector<16) return 0; else {
      u16t cs64 = GetCodeSelector();
      int    ii;

      IntsOff();
      lapic    = apic;
      SavedSpr = apic[APIC_SPURIOUS];

      for (ii=0; ii<2; ii++) {
         u32t   vn = ii?spr_vector:tmr_vector;
         u64t inta = (u64t)(ii?&iretq_addr:&timer_irq);
         struct gate64_s *idesc = (struct gate64_s *)CurIDT.base + vn;
         idesc->g64_sel    = cs64;
         idesc->g64_ofs0   = (u16t)inta;
         idesc->g64_ofs16  = (u16t)(inta>>16);
         idesc->g64_ofs32  = (u32t)(inta>>32);
         idesc->g64_parms  = 0;
         idesc->g64_access = D_PRES|D_INTGATE32;
         // mark it to restore
         int_used[vn] = 1;
      }
      IntsOn();

      return 1;
   }
}

/* Function replaces EFI allocated block for GDT, IDT backup & stack to
   the buffer, provided by 32-bit part (in QSINIT memory area).

   Called immediately after ExitBootServices() to make sure, that GDT will
   stay valid when all EfiLoaderData memory will be marked as "free".

   Buffer is 128k, this should be enough (max 64k GDT + 4k IDT + stack) */
void ReplaceGDT(void *buffer) {
   if (GDTCopyPages) {
      EFI_PHYSICAL_ADDRESS newGDT = (EFI_PHYSICAL_ADDRESS)buffer;
      struct gate64_s *    newIDT = (struct gate64_s *)(newGDT + GDTSize);
      // disable ints to be safe
      IntsOff();
      RtCopyMem(buffer, (void*)GDTCopy, GDTSize);
      RtCopyMem(newIDT, SavedIDTv, sizeof(struct gate64_s)*256);

      Header32->gdt_pos = (u32t)newGDT;

      XcptStack    = (u8t*)newGDT + _128KB;
      SavedIDTv    = newIDT;
      GDTCopy      = newGDT;

      CurGDT.base  = GDTCopy;
      CurGDT.limit = GDTSize - 1;

      WriteGDTR(&CurGDT);

      IntsOn();
      // free memory is just not possible, because no boot services ...
      GDTCopyPages = 0;
   }
}

void RestoreIDT(void) {
   struct gate64_s *idesc = (struct gate64_s *)CurIDT.base;
   int  ii;

   IntsOff();
   for (ii=0; ii<256; ii++)
      if (int_used[ii]) idesc[ii] = SavedIDTv[ii];
   IntsOn();
}

u32t LoadExitModule(char *module, char *args, u32t modtype) {
   EFI_STATUS              st = EFI_SUCCESS;
   UINTN               arglen = 0;
   EFI_LOADED_IMAGE *RunImage = 0;
   u32t                   err = 0;
   CHAR16                 *fp = 0;

   if (RunModHandle) return 0x0001E; // E_SYS_INITED
   if (args) arglen = strlen((char*)args);

   if (modtype & EFBM_MEMORY) {
      st = BS->LoadImage(FALSE, SelfImg, 0, module, modtype&EFBM_SIZEMASK,
                         &RunModHandle);
      // name is unknown, we just need something for argv[0]
      if (!EFI_ERROR(st)) fp = strto16("0", 0);
   } else {
      fp = strto16(module, 0);
      if (fp) {
         // current dir?
         if (*module!='\\') {
            CHAR16 *fpc = (CHAR16*)AllocatePool(StrSize(fp) + StrSize(DevPath) + 2);
            if (fpc) {
               wcscpy(fpc, DevPath);
               StrCat(fpc, L"\\");
               StrCat(fpc, fp);
               FreePool(fp);
               fp = fpc;
            } else
               st = EFI_OUT_OF_RESOURCES;
         }
         if (!EFI_ERROR(st)) {
            EFI_DEVICE_PATH *dp = FileDevicePath(LoadedImage->DeviceHandle, fp);
            if (!dp) st = EFI_OUT_OF_RESOURCES; else {
               st = BS->LoadImage(FALSE, SelfImg, dp, 0, 0, &RunModHandle);
               FreePool(dp);
            }
         }
      }
   }
   if (!fp && !EFI_ERROR(st)) st = EFI_OUT_OF_RESOURCES;

   if (!EFI_ERROR(st) && arglen)
      st = BS->HandleProtocol(RunModHandle, &LoadedImageProtocol, (VOID*)&RunImage);

   if (!EFI_ERROR(st)) {
      // setup parameters
      if (arglen) {
         CHAR16 *lp;
         RunModOpt = (CHAR16*)AllocatePool(StrSize(fp) + arglen*2 + 4);
         wcscpy(RunModOpt, fp);
         lp    = RunModOpt + wcslen(RunModOpt);
         *lp++ = ' ';
         while (*args) *lp++ = *args++;
         *lp   = 0;
         RunImage->LoadOptionsSize  = (UINT32)StrSize(RunModOpt);
         RunImage->LoadOptions      = (VOID*)RunModOpt;
      }
   } else {
      if (RunModHandle) {
         BS->UnloadImage(RunModHandle);
         RunModHandle = 0;
      }
      switch (st) {
         case EFI_NOT_FOUND         : err = 0x0000F; break; // E_SYS_ZEROPTR
         case EFI_INVALID_PARAMETER : err = 0x00008; break; // E_SYS_INVPARM
         case EFI_UNSUPPORTED       : err = 0x0001F; break; // E_SYS_BADFMT
         case EFI_OUT_OF_RESOURCES  : err = 0x00002; break; // E_SYS_NOMEM
         case EFI_LOAD_ERROR        : err = 0x10005; break; // E_MOD_BROKENFILE
         case EFI_DEVICE_ERROR      : err = 0x20001; break; // E_DSK_ERRREAD
         case EFI_ACCESS_DENIED     : err = 0x00007; break; // E_SYS_ACCESS
         case EFI_SECURITY_VIOLATION: err = 0x10039; break; // E_MOD_EFISEC
         default:
            err = 0x00020; // E_SYS_SOFTFAULT
      }

   }
   if (fp) FreePool(fp); 
   return err;
}

// "main" function
void QSLoader(IN EFI_HANDLE ImgHandle, IN EFI_SYSTEM_TABLE *SysTab) {
   EFI_STATUS   st;
   UINTN        ii;
   // Initialize the Library. Set BS, RT, &ST globals
   //  BS = Boot Services RT = RunTime Services
   //  ST = System Table
   InitializeLib (ImgHandle, SysTab);

   SelfImg = ImgHandle;

   while (1) {
      // save and replace GDT 
      ReadGDTR(&SavedGDT);
      
      if (0xFFFF - SavedGDT.limit < RESERVE_DESCRIPTORS * 8)
         ErrorExit(L"There is no space in System GDT, unable to run!\n");
      // who can allocate above 4Gb? still care
      GDTSize      = SavedGDT.limit + 1 + RESERVE_DESCRIPTORS * 8;
      if (GDTSize>_64KB) GDTSize = _64KB;
      GDTCopyPages = Round4k(GDTSize + sizeof(struct gate64_s)*256 +
                             XCPT_STACK_SIZE)>>12;
      GDTCopy      = PageAlloc(GDTCopyPages);
      // makes a copy of original GDT
      RtZeroMem((void*)GDTCopy, GDTSize);
      RtCopyMem((void*)GDTCopy, (void*)SavedGDT.base, GDTLowest = SavedGDT.limit + 1);
      
      RtZeroMem(&int_used, sizeof(int_used));
      
      SavedIDTv    = (struct gate64_s *)(GDTCopy + GDTSize);
      // points to the end of block
      XcptStack    = (u8t*)(SavedIDTv + 256) + XCPT_STACK_SIZE;
      
      CurGDT.base  = GDTCopy;
      CurGDT.limit = GDTSize - 1;
      
      SaveSegRegs();
      WriteGDTR(&CurGDT);
#if PRINT_MEM_TABLE > 0
      Print(L"GDT base: %X, size %X\n", GDTCopy, GDTSize);
#endif
      BS->SetWatchdogTimer(0, 0, 0, 0);
      
      // Get the device handle and file path to the EFI OS Loader itself.
      st = BS->HandleProtocol(ImgHandle, &LoadedImageProtocol, (VOID*)&LoadedImage);
      if (EFI_ERROR(st))
         ErrorExit(L"Can not retrieve a LoadedImageProtocol handle for ImageHandle\n");
      
      st = BS->HandleProtocol(LoadedImage->DeviceHandle, &DevicePathProtocol, (VOID*)&DevicePath);
      if (EFI_ERROR(st) || DevicePath==NULL)
         ErrorExit(L"Can not find a DevicePath handle for LoadedImage\n");
      // command line parser
      parse_args();
      // init keyboard
      key_init();
      // init console
      vio_init();
      // calculate "timer" ticks
      tm_counter();
      // find a place to place 32-bit part ;)
      init_memory();
#if PRINT_MEM_TABLE > 0
      Print(L"  QSINIT memory at address   : %X\n", QSMem);
      Print(L"         size in kb          : %ld\n", ((u64t)QSPages<<PAGESHIFT)>>10);

      Print(L"\nPress ENTER to continue...\n");
      {
         u32t key;
         do {
            key = key_read();
            Print (L"Key Code : %X\n", key);
         } while ((key&0xFF) != 13);
      } 
#endif
      if (!SafeMode) {
         // check for a key at least once and hope it will return LEFT SHIFT for us ;)
         key_pressed();
         if (key_status()&KEY_SHIFTLEFT) SafeMode = 1;
      }
      // save IDT
      ReadIDTR(&CurIDT);
      RtCopyMem(SavedIDTv, (void*)CurIDT.base, sizeof(struct gate64_s)*256);
      // save CR4
      SavedCr4 = hlp_setreg(4, FFFF64);
   
      ii = Load32();
      if (ii) Print(L"Load32() rc = %d\n",ii);
      ii = SysConfig();
      if (ii) Print(L"SysConfig() rc = %d\n",ii);
      ii = VioConfig();
      if (ii) Print(L"VioConfig() rc = %d\n",ii);
      ii = Exec32();
      // restore APIC
      if (lapic) {
         IntsOff();
         lapic[APIC_LVT_TMR]  = APIC_DISABLE;
         lapic[APIC_SPURIOUS] = SavedSpr;
         IntsOn();
      }
      if (ii) {
         // these codes are really hardcoded in many places ...
         if (ii==4 || ii>=10 && ii<=14) {
            Print(L"\nFatal: ");
            switch (ii) {
               case  4: Print(L"Software fault occurs!\n"); break;
               case 10: Print(L"Memory allocation error!\n"); break;
               case 11: Print(L"FAT processing error\n"); break;
               case 12: Print(L"Disk read error!\n"); break;
               case 13: Print(L"%a is missing or invalid.\n", LdiName); break;
               case 14: Print(L"Unable to unpack the secondary code.\n"); break;
            }
            Print(L"Press any key to boot next device.\n"); break;
            if (key_pressed()) key_read();
            key_read();
         } else
            Print(L"Exec32() rc = %d\n",ii);
      }
      RestoreIDT();
      // restore cr4
      hlp_setreg(4, SavedCr4);
      // free resources
      Shutdown(RunModHandle ? 0 : L"Bye!\n");

      if (RunModHandle) {
         BS->StartImage(RunModHandle, 0, 0); 
         if (RunModOpt) {
            FreePool(RunModOpt);
            RunModOpt = 0;
         }
         RunModHandle = 0;
         continue;
      }
      break;
   }
   // return to EFI shell/boot sequence
   BS->Exit(SelfImg, EFI_SUCCESS, 0, NULL);
}
